from typing import Any, Type, TypeVar, Optional, Union, get_origin, get_args
from enum import Enum
from pydantic import (
    BaseModel,
    ConfigDict,
    ValidationInfo,
    computed_field,
    field_validator,
)
import math
import numpy as np


def serialize(data: Any):
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = serialize(value)
        return data
    elif isinstance(data, list):
        if len(data) == 1:
            return serialize(data[0])
        return [serialize(item) for item in data]
    elif isinstance(data, np.floating):
        return float(data)
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, np.bool_):
        return bool(data)
    elif isinstance(data, np.ndarray):
        return serialize(np.squeeze(data).tolist())
    elif isinstance(data, TaggedModel):
        return data.model_dump()
    if type(data).__module__ == np.__name__:
        # final catchall
        data = np.squeeze(data).tolist()
    else:
        return data

def _allows_none(tp: Any) -> bool:
    """
    Return True iff the type annotation *explicitly* admits `None`.
    Handles Optional[T], Union[..., None], and Annotated[…].
    """
    if tp is None or tp is type(None):
        return True

    origin = get_origin(tp)
    if origin is Union:  # Optional[T] is just Union[T, None]
        return type(None) in get_args(tp)
    if origin is getattr(__import__("typing"), "Annotated", None):
        # drill into Annotated[T, ...]  (first arg is the real annotation)
        return _allows_none(get_args(tp)[0])

    return False

class TaggedModel(BaseModel):
    @computed_field(repr=False)
    @property
    def __data_type__(self) -> str:
        return self.__class__.__name__

    @field_validator("*", mode="before")
    @classmethod
    def _nan_to_none(cls, v: Any, info: ValidationInfo):
        # this is necessary because matlab doesn't support None, so they are all
        # changed to NaNs in serialization.
        if not (isinstance(v, float) and math.isnan(v)):
            return v  # nothing to do
        if info.field_name is None:
            # this should never happen...
            return v

        field_type = cls.model_fields[info.field_name].annotation
        if _allows_none(field_type):
            return None  # safe to coerce
        return v  # leave `nan` as-is

class LicenseType(Enum):
    _2D = "2d"
    _3D = "3d"
    default = "default"

    def to_dict(self):
        return {"__type__": "LicenseType", "value": self.value}

    def __str__(self):
        return self.value

TensorType = Union[float, list[float], list[list[float]]]
DTensorType = Union[list[list[float]]]

class MaterialProperties(TaggedModel):
    n: Optional[TensorType] = None
    eps: Optional[TensorType] = None
    mu: Optional[TensorType] = None
    d: Optional[DTensorType] = None

    # this is necessary to support np.ndarrays here...
    model_config = ConfigDict(arbitrary_types_allowed=True)

class MaterialSpec(TaggedModel):
    material: Union[str, MaterialProperties]
    theta: Optional[float] = None
    phi: Optional[float] = None
    x: Optional[float] = None
    loss: Optional[float] = None  # dB/m

class Grid(TaggedModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    x: np.ndarray
    y: np.ndarray

    is_pml: bool = False
    is_expanded: bool = False
    is_bc: bool = False

class Field(TaggedModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    field: np.ndarray
    field_names: list[str]
    num_modes: int
    grid: Grid

T = TypeVar("T")

_TYPE_REGISTRY = {
    "MaterialSpec": MaterialSpec,
    "MaterialProperties": MaterialProperties,
    "Field": Field,
    "Grid": Grid,
}

def register_type(cls: Type[T]) -> Type[T]:
    """Class decorator to register an Exception subclass."""
    name = cls.__name__
    if name in _TYPE_REGISTRY:
        raise ValueError(f"{name} already registered")
    _TYPE_REGISTRY[name] = cls
    return cls

def get_type(name: str) -> Type:
    try:
        return _TYPE_REGISTRY[name]
    except KeyError:
        raise KeyError(f"Unknown exception type: {name}")

def object_from_dict(data: dict[str, Any]) -> Any:
    """
    Reconstructs an exception from its serialized form.
    Expects data["__data_type__"] to be the class name.
    """
    name = data.pop("__data_type__")
    if not name:
        raise KeyError("Missing 'type' in exception data")

    ExcClass = get_type(name)

    return ExcClass(**data)

@register_type
class EModeError(Exception):
    def __init__(self, msg: str = '', *args, **kwargs):
        super().__init__(msg, *args, **kwargs)
        self._custom_fields = []
        self.msg = msg

    def __setattr__(self, name: str, value: Any) -> None:
        if name != "_custom_fields":
            self._custom_fields.append(name)
        return super().__setattr__(name, value)

    def to_dict(self) -> dict:
        d: dict[str, Any] = {
            "__data_type__": self.__class__.__name__,
        }
        for f in self._custom_fields:
            d[f] = getattr(self, f, None)

        return d

@register_type
class ArgumentError(EModeError):
    def __init__(self, msg: str, function: Optional[str], argument: Optional[str]):
        super().__init__(msg)
        self.msg = msg
        self.function = function
        self.argument = argument

    def __str__(self):
        return f'ArgumentError: the argument: ({self.argument}) to function: ({self.function}) had error: "{self.msg}"'

@register_type
class EPHKeyError(EModeError):
    def __init__(self, msg: str, filename: str, key: str):
        super().__init__(msg)
        self.msg = msg
        self.filename = filename
        self.key = key

    def __str__(self):
        return f'EPHKeyError: the key: ({self.key}) doesn\'t exist in the file: ({self.filename}), "{self.msg}"'

@register_type
class FileError(EModeError):
    def __init__(self, msg: str, filename: str):
        super().__init__(msg)
        self.msg = msg
        self.filename = filename

    def __str__(self):
        return f'FileError: the file: "{self.filename}" had error: "{self.msg}"'

@register_type
class LicenseError(EModeError):
    def __init__(self, msg: str, license_type: LicenseType | None):
        super().__init__(msg, license_type)
        self.msg = msg
        self.license_type = license_type

    def __str__(self):
        if isinstance(self.license_type, LicenseType):
            lt = self.license_type['value'].lower()  # type: ignore
        else:
            lt = str(self.license_type)  # guard against None
        if lt == '3d':
            emLicense = 'EMode3D'
        else:
            emLicense = 'EMode2D'
        return f'LicenseError: current license: "{emLicense}", error msg: "{self.msg}"'

@register_type
class ShapeError(EModeError):
    def __init__(self, msg: str, shape_name: str):
        super().__init__(msg)
        self.msg = msg
        self.shape_name = shape_name

    def __str__(self):
        return f'ShapeError: error: "{self.msg}" with shape: {self.shape_name}'

@register_type
class NameError(EModeError):
    def __init__(self, msg: str, type: str, name: str):
        super().__init__(msg)
        self.msg = msg
        self.type = type
        self.name = name

    def __str__(self):
        return f'NameError: error: "{self.msg}" for type: {self.type} and name: {self.name}'

@register_type
class NotImplementedError(EModeError):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return f"NotImplementedError: {self.msg}"



import atexit
from datetime import datetime, timezone
from logging import getLogger
from logging.config import dictConfig
import os
from pathlib import Path
import platform
from subprocess import Popen, PIPE
from threading import Thread
import time
from typing import Optional, Literal, Union
from .emodeclient import EModeClient
from .file_utils import Cache
from .types import EModeError

logger = getLogger(__name__)


def _forward_stdout(pipe):
    try:
        for line in iter(pipe.readline, ""):
            # print() => IPython captures it into the cell output
            print(line, end="", flush=True)
    finally:
        pipe.close()

class EMode:
    def __init__(
        self,
        sim: Optional[str] = None,
        simulation_name: Optional[str] = "emode",
        license_type: Literal["2d", "3d", "default"] = "default",
        save_path: Union[str, Path] = ".",
        verbose: bool = False,
        roaming: bool = False,
        open_existing: bool = False,
        new_name: Union[bool, str] = False,
        priority: Literal["pH", "pAN", "pBN", "pI", "pN"] = "pN",
        emode_cmd: Optional[list[str]] = None,
    ):
        """
        Initialize defaults and create an EMode session.

        parameters:
        -----
        simulation_name: str
            The name of the default simulation to load.

        sim: str
            Alias for simulation_name.  simulation_name takes precedence.

        license_type: Literal['2d','3d','default'] = 'default'
            The type of license you wish to check out for this session.

        save_path: str | Path = '.'
            The path to save results.

        verbose: bool = False
            Verbose output from EMode.

        roaming: bool = False
            Enable roaming mode

        open_existing: bool = False
            open an existing simulation.  If False, and <save_path>/<simulation_name>.eph
            exists, <save_path>/<simulation_name>_0.eph will be created.

        new_name: bool = False
            I'm not sure what this does?

        priority: Literal['pH', 'pAN', 'pN', 'pBN', 'pI'] = 'pN'
            The EMode process priority:
                'pH': High priority
                'pAN': Above Normal priority
                'pN': Normal priority
                'pBN': Below Normal priority
                'pI': Idle priority

        emode_cmd: Optional[list[str]] = None
            The command to use to invoke EMode as a command list.  This shouldn't need to be
            modified if you installed EMode normally.
        """
        self.setup_logging()

        if sim:
            logger.warning("The `sim` argument in the `EMode` class is depreciated, use `simulation_name` instead.")
        
        simulation_name = simulation_name or sim

        if not isinstance(simulation_name, str):
            raise TypeError("parameter 'simulation_name' must be a string")

        if not isinstance(save_path, (str, Path)):
            raise TypeError("parameter 'save_path' must be a string or pathlib.Path")

        if license_type not in ["2d", "3d", "default"]:
            raise ValueError(
                "parameter 'license_type' must be one of ['2d','3d','default']"
            )

        if priority not in ["pH", "pAN", "pN", "pBN", "pI"]:
            raise ValueError(
                "parameter 'priority' must be one of ['pH','pAN','pN','pBN','pI']"
            )

        self.dsim = simulation_name
        self.priority = priority
        self.verbose = verbose
        self.license_type = license_type
        self.roaming = roaming
        self.ext = ".eph"
        self.port_file_label = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S%f")
        self.cache = Cache(self.port_file_label)

        if self.in_ipython():
            self.proc = Popen(
                self.build_cmd_list(emode_cmd),
                stdout=PIPE,
                stderr=None,
                text=True,
                bufsize=1,
            )
            self.setup_print_thread()
        else:
            self.proc = Popen(
                self.build_cmd_list(emode_cmd),
                stdout=None,
                stderr=None,
                # stderr=None,
            )
        self.client = EModeClient(self.cache)
        atexit.register(self.close_atexit)

        if open_existing:
            RV = self.call(
                "EM_open", simulation_name=simulation_name, save_path=save_path, new_simulation_name=new_name
            )
        else:
            RV = self.call("EM_init", simulation_name=simulation_name, save_path=save_path)

        if RV == "failed":
            raise EModeError("EMode failed to launch.")
        self.dsim = RV[len("sim:") :]

    def build_cmd_list(self, emode_cmd):
        if platform.system() == "Windows":
            cmd = emode_cmd or ["EMode.exe"]
        else:
            cmd = emode_cmd or ["emode"]

        cmd += ["run", self.port_file_label]
        if self.license_type != "default":
            cmd += [f"-{self.license_type}"]
        if self.verbose:
            cmd += ["-v"]
        if self.priority != "pN":
            cmd += [self.priority]
        if self.roaming:
            cmd += ["-r"]

        logger.info(f"emode command list: {cmd}")
        self.cmd = cmd
        return cmd

    def setup_logging(self):
        if log_level := os.getenv("EMODE_LOGGING", "WARNING"):
            dictConfig(
                {
                    "version": 1,
                    "disable_existing_loggers": False,
                    "formatters": {
                        "std": {
                            "format": "%(asctime)s.%(msecs)03d %(name)s %(levelname)s %(funcName)s %(message)s"
                        }
                    },
                    "handlers": {
                        "console": {
                            "class": "logging.StreamHandler",
                            "formatter": "std",
                            "level": log_level,
                        }
                    },
                    "loggers": {
                        "": {"handlers": ["console"], "level": log_level},
                        "matplotlib": {"level": "WARNING"},
                        "PIL": {"level": "WARNING"},
                    },
                }
            )

    def setup_print_thread(self):
        # self.print_thread = Thread(
        #     target=lambda: shutil.copyfileobj(
        #         self.proc.stdout, sys.stdout.buffer, length=1
        #     ),
        #     daemon=True,
        # )
        self.print_thread = Thread(
            name="print EMode output",
            target=_forward_stdout,
            args=(self.proc.stdout,),
            daemon=True,
        )
        self.print_thread.start()

    def in_ipython(self):
        try:
            _ = get_ipython()  # type: ignore
            self.ipython = True
        except NameError:
            self.ipython = False

        return self.ipython

    def call(self, function: str, **kwargs) -> Union[dict, Literal['failed'], float, str, list]:
        logger.debug(f"calling '{function}' with args: {kwargs}")
        if not isinstance(function, str):
            raise TypeError("parameter 'function' must be of type 'str'")

        sendset = kwargs
        sendset.update({"function": function})

        if "sim" not in sendset and "simulation_name" not in sendset:
            sendset["simulation_name"] = self.dsim

        self.client.send(sendset)

        try:
            return self.client.recv()
        except ConnectionResetError:
            logger.debug("connection reset from EMode, shutting down")
            self.client.close()
            return "failed"

    def close(self, **kwargs):
        logger.debug(f"closing connection with kwargs {kwargs}")
        try:
            self.call("EM_close", **kwargs)
            self.client.close()

            while True:
                time.sleep(0.01)
                if self.proc.poll() is None or self.proc.poll() is self.proc.returncode:
                    break
                time.sleep(0.01)

        except Exception:
            logger.exception("got exception closing client")

        return

    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            if args:
                kwargs["key"] = args[0]
                if len(args) > 1:
                    raise ValueError("Please pass all arguments as kwargs")
            return self.call("EM_" + name, **kwargs)

        return wrapper

    def close_atexit(self):
        if self.client.connected:
            self.close()

        return



from .emodeconnection import EMode
from .eph_utils import open_file, get, inspect
from .types import MaterialSpec, MaterialProperties
from subprocess import Popen
from platform import system

__all__ = ["EMode", "open_file", "get", "inspect", "MaterialSpec", "MaterialProperties", "EModeLogin"]


def EModeLogin():
    emodecmd = "emode"
    if system() == "Windows":
        emodecmd = "EMode.exe"
    Popen([emodecmd])



import os
import pickle
import numpy as np
import scipy.io as sio


def open_file(sim="emode", simulation_name=None):
    """
    Opens an EMode simulation file with either .eph or .mat extension.
    """
    sim = simulation_name or sim
    if not isinstance(sim, str):
        raise TypeError("input parameter 'simulation_name' must be a string")

    ext = ".eph"
    mat = ".mat"
    found = False
    for file in os.listdir():
        if (file == sim + ext) or ((file == sim) and (sim.endswith(ext))):
            found = "eph"
            with open(file, "rb") as fl:
                f = pickle.load(fl)
        elif (file == sim + mat) or ((file == sim) and (sim.endswith(mat))):
            found = "mat"
            f = loadmat(sim + mat)

    if not found:
        raise FileNotFoundError(
            f"no file: {sim}, {sim + ext}, or {sim + mat} was found"
        )

    return f


def get(variable, sim="emode", simulation_name=None):
    """
    Return data from simulation file.
    """
    if not isinstance(variable, str):
        raise TypeError("input parameter 'variable' must be a string")

    sim = simulation_name or sim
    if not isinstance(sim, str):
        raise TypeError("input parameter 'simulation_name' must be a string")

    f = open_file(sim=sim)

    if variable in list(f.keys()):
        data = f[variable]
    else:
        print("Data does not exist.")
        return

    return data


def inspect(sim="emode", simulation_name=None):
    """
    Return list of keys from available data in simulation file.
    """
    sim = simulation_name or sim
    if not isinstance(sim, str):
        raise TypeError("input parameter 'simulation_name' must be a string")

    f = open_file(sim=sim)

    fkeys = list(f.keys())
    fkeys.remove("EMode_simulation_file")
    return fkeys


def loadmat(filename):
    data = sio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


def _check_keys(data):
    force_1D = ["effective_index", "TE_indices", "TM_indices"]
    force_3D = ["Fx", "Fy", "Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
    for key in data:
        if isinstance(data[key], sio.matlab.mio5_params.mat_struct):  # type: ignore
            data[key] = _todict(data[key])

        if key in force_1D:
            if (not isinstance(data[key], list)) and (
                not isinstance(data[key], np.ndarray)
            ):
                data[key] = np.expand_dims(data[key], axis=0)

        if key in force_3D:
            if len(np.shape(data[key])) < 3:
                data[key] = np.expand_dims(data[key], axis=0)

    return data


def _todict(matobj):
    ddict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, sio.matlab.mio5_params.mat_struct):  # type: ignore
            ddict[strg] = _todict(elem)
        if isinstance(elem, str):
            ddict[strg] = elem.strip()
        else:
            ddict[strg] = elem
    return ddict
